"""
Spectral Distributions of Filters
=================================

Defines the spectral distributions of filters.

Each filter data is in the form of :class:`dict` class instance of
:class:`colour.SpectralDistribution` classes as follows::

    {'name': SpectralDistribution, ..., 'name': SpectralDistribution}

The following filters are available:

-   ISO 7589 Diffuser

References
----------
-   :cite:`InternationalOrganizationforStandardization2002` : International
    Organization for Standardization. (2002). INTERNATIONAL STANDARD ISO
    7589-2002 - Photography - Illuminants for sensitometry - Specifications for
    daylight, incandescent tungsten and printer.
"""

from __future__ import annotations

from functools import partial

from colour.colorimetry import SpectralDistribution
from colour.hints import Dict
from colour.utilities import LazyCaseInsensitiveMapping

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "DATA_FILTERS_ISO",
    "SDS_FILTERS_ISO",
    "SDS_FILTERS",
]

DATA_FILTERS_ISO: Dict = {
    "ISO 7589 Diffuser": {
        350: 0.00,
        360: 0.00,
        370: 0.00,
        380: 0.10,
        390: 0.43,
        400: 0.69,
        410: 0.78,
        420: 0.83,
        430: 0.86,
        440: 0.88,
        450: 0.90,
        460: 0.91,
        470: 0.93,
        480: 0.94,
        490: 0.95,
        500: 0.96,
        510: 0.97,
        520: 0.98,
        530: 0.99,
        540: 0.99,
        550: 1.00,
        560: 1.00,
    }
}

SDS_FILTERS_ISO: LazyCaseInsensitiveMapping = LazyCaseInsensitiveMapping(
    {
        "ISO 7589 Diffuser": partial(
            SpectralDistribution,
            DATA_FILTERS_ISO["ISO 7589 Diffuser"],
            name="ISO 7589 Diffuser",
        ),
    }
)
SDS_FILTERS_ISO.__doc__ = """
Spectral distributions of *ISO* filters.

References
----------
:cite:`InternationalOrganizationforStandardization2002`
"""

SDS_FILTERS: LazyCaseInsensitiveMapping = LazyCaseInsensitiveMapping(
    SDS_FILTERS_ISO
)
SDS_FILTERS.__doc__ = """
Spectral distributions of filters.

References
----------
:cite:`InternationalOrganizationforStandardization2002`
"""



"""
CIE Illuminant D Series Correlated Colour Temperature
=====================================================

Defines the *CIE Illuminant D Series* correlated colour temperature
:math:`T_{cp} computations objects:

-   :func:`colour.temperature.xy_to_CCT_CIE_D`: Correlated colour temperature
    :math:`T_{cp}` computation of a *CIE Illuminant D Series* from its *CIE xy*
    chromaticity coordinates.
-   :func:`colour.temperature.CCT_to_xy_CIE_D`: *CIE xy* chromaticity
    coordinates computation of a *CIE Illuminant D Series* from its correlated
    colour temperature :math:`T_{cp}`.

References
----------
-   :cite:`Wyszecki2000z` : Wyszecki, Günther, & Stiles, W. S. (2000). CIE
    Method of Calculating D-Illuminants. In Color Science: Concepts and
    Methods, Quantitative Data and Formulae (pp. 145-146). Wiley.
    ISBN:978-0-471-39918-6
"""

from __future__ import annotations

import numpy as np
from scipy.optimize import minimize

from colour.colorimetry import daylight_locus_function
from colour.hints import (
    ArrayLike,
    Dict,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Optional,
)
from colour.utilities import as_float_array, as_float, tstack, usage_warning

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "xy_to_CCT_CIE_D",
    "CCT_to_xy_CIE_D",
]


def xy_to_CCT_CIE_D(
    xy: ArrayLike, optimisation_kwargs: Optional[Dict] = None
) -> FloatingOrNDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` of a
    *CIE Illuminant D Series* from its *CIE xy* chromaticity coordinates.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates.
    optimisation_kwargs
        Parameters for :func:`scipy.optimize.minimize` definition.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`.

    Warnings
    --------
    The *CIE Illuminant D Series* method does not give an analytical inverse
    transformation to compute the correlated colour temperature :math:`T_{cp}`
    from given *CIE xy* chromaticity coordinates, the current implementation
    relies on optimization using :func:`scipy.optimize.minimize` definition and
    thus has reduced precision and poor performance.

    References
    ----------
    :cite:`Wyszecki2000z`

    Examples
    --------
    >>> xy_to_CCT_CIE_D(np.array([0.31270775, 0.32911283]))
    ... # doctest: +ELLIPSIS
    6504.3895840...
    """

    xy = as_float_array(xy)
    shape = xy.shape
    xy = np.atleast_1d(xy.reshape([-1, 2]))

    def objective_function(
        CCT: FloatingOrArrayLike, xy: ArrayLike
    ) -> FloatingOrNDArray:
        """Objective function."""

        objective = np.linalg.norm(CCT_to_xy_CIE_D(CCT) - xy)

        return as_float(objective)

    optimisation_settings = {
        "method": "Nelder-Mead",
        "options": {
            "fatol": 1e-10,
        },
    }
    if optimisation_kwargs is not None:
        optimisation_settings.update(optimisation_kwargs)

    CCT = as_float_array(
        [
            minimize(
                objective_function,
                x0=6500,
                args=(xy_i,),
                **optimisation_settings,
            ).x
            for xy_i in as_float_array(xy)
        ]
    )

    return as_float(np.reshape(CCT, shape[:-1]))


def CCT_to_xy_CIE_D(CCT: FloatingOrArrayLike) -> NDArray:
    """
    Return the *CIE xy* chromaticity coordinates of a
    *CIE Illuminant D Series* from its correlated colour temperature
    :math:`T_{cp}`.

    Parameters
    ----------
    CCT
        Correlated colour temperature :math:`T_{cp}`.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    Raises
    ------
    ValueError
        If the correlated colour temperature is not in appropriate domain.

    References
    ----------
    :cite:`Wyszecki2000z`

    Examples
    --------
    >>> CCT_to_xy_CIE_D(6504.38938305)  # doctest: +ELLIPSIS
    array([ 0.3127077...,  0.3291128...])
    """

    CCT = as_float_array(CCT)

    if np.any(CCT[np.asarray(np.logical_or(CCT < 4000, CCT > 25000))]):
        usage_warning(
            "Correlated colour temperature must be in domain "
            "[4000, 25000], unpredictable results may occur!"
        )

    CCT_3 = CCT**3
    CCT_2 = CCT**2

    x = np.where(
        CCT <= 7000,
        -4.607 * 10**9 / CCT_3
        + 2.9678 * 10**6 / CCT_2
        + 0.09911 * 10**3 / CCT
        + 0.244063,
        -2.0064 * 10**9 / CCT_3
        + 1.9018 * 10**6 / CCT_2
        + 0.24748 * 10**3 / CCT
        + 0.23704,
    )

    y = daylight_locus_function(x)

    return tstack([x, y])



"""
ITU-R BT.2020
=============

Defines the *ITU-R BT.2020* opto-electrical transfer function (OETF) and
electro-optical transfer function (EOTF):

-   :func:`colour.models.eotf_inverse_BT2020`
-   :func:`colour.models.eotf_BT2020`

References
----------
-   :cite:`InternationalTelecommunicationUnion2015h` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.2020 - Parameter
    values for ultra-high definition television systems for production and
    international programme exchange (pp. 1-8).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2020-2-201510-I!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.hints import Boolean, FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import (
    Structure,
    as_float,
    domain_range_scale,
    from_range_1,
    to_domain_1,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_BT2020",
    "CONSTANTS_BT2020_PRECISE",
    "eotf_inverse_BT2020",
    "eotf_BT2020",
]

CONSTANTS_BT2020: Structure = Structure(
    alpha=lambda x: 1.0993 if x else 1.099,
    beta=lambda x: 0.0181 if x else 0.018,
)
"""*BT.2020* colourspace constants."""

CONSTANTS_BT2020_PRECISE: Structure = Structure(
    alpha=lambda x: 1.09929682680944, beta=lambda x: 0.018053968510807
)
"""
*BT.2020* colourspace constants at double precision to connect the two curve
segments smoothly.

References
----------
:cite:`InternationalTelecommunicationUnion2015h`
"""


def eotf_inverse_BT2020(
    E: FloatingOrArrayLike,
    is_12_bits_system: Boolean = False,
    constants: Structure = CONSTANTS_BT2020,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2020* inverse electro-optical transfer
    function (EOTF).

    Parameters
    ----------
    E
        Voltage :math:`E` normalised by the reference white level and
        proportional to the implicit light intensity that would be detected
        with a reference camera colour channel R, G, B.
    is_12_bits_system
        *BT.709* *alpha* and *beta* constants are used if system is not 12-bit.
    constants
        *Recommendation ITU-R BT.2020* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Resulting non-linear signal :math:`E'`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015h`

    Examples
    --------
    >>> eotf_inverse_BT2020(0.18)  # doctest: +ELLIPSIS
    0.4090077...
    """

    E = to_domain_1(E)

    a = constants.alpha(is_12_bits_system)
    b = constants.beta(is_12_bits_system)

    E_p = np.where(E < b, E * 4.5, a * spow(E, 0.45) - (a - 1))

    return as_float(from_range_1(E_p))


def eotf_BT2020(
    E_p: FloatingOrArrayLike,
    is_12_bits_system: Boolean = False,
    constants: Structure = CONSTANTS_BT2020,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2020* electro-optical transfer function
    (EOTF).

    Parameters
    ----------
    E_p
        Non-linear signal :math:`E'`.
    is_12_bits_system
        *BT.709* *alpha* and *beta* constants are used if system is not 12-bit.
    constants
        *Recommendation ITU-R BT.2020* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Resulting voltage :math:`E`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015h`

    Examples
    --------
    >>> eotf_BT2020(0.705515089922121)  # doctest: +ELLIPSIS
    0.4999999...
    """

    E_p = to_domain_1(E_p)

    a = constants.alpha(is_12_bits_system)
    b = constants.beta(is_12_bits_system)

    with domain_range_scale("ignore"):
        E = np.where(
            E_p < eotf_inverse_BT2020(b),
            E_p / 4.5,
            spow((E_p + (a - 1)) / a, 1 / 0.45),
        )

    return as_float(from_range_1(E))



"""Showcases colour spectrum computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box("Spectrum Computations")

data_sample = {
    380: 0.048,
    385: 0.051,
    390: 0.055,
    395: 0.060,
    400: 0.065,
    405: 0.068,
    410: 0.068,
    415: 0.067,
    420: 0.064,
    425: 0.062,
    430: 0.059,
    435: 0.057,
    440: 0.055,
    445: 0.054,
    450: 0.053,
    455: 0.053,
    460: 0.052,
    465: 0.052,
    470: 0.052,
    475: 0.053,
    480: 0.054,
    485: 0.055,
    490: 0.057,
    495: 0.059,
    500: 0.061,
    505: 0.062,
    510: 0.065,
    515: 0.067,
    520: 0.070,
    525: 0.072,
    530: 0.074,
    535: 0.075,
    540: 0.076,
    545: 0.078,
    550: 0.079,
    555: 0.082,
    560: 0.087,
    565: 0.092,
    570: 0.100,
    575: 0.107,
    580: 0.115,
    585: 0.122,
    590: 0.129,
    595: 0.134,
    600: 0.138,
    605: 0.142,
    610: 0.146,
    615: 0.150,
    620: 0.154,
    625: 0.158,
    630: 0.163,
    635: 0.167,
    640: 0.173,
    645: 0.180,
    650: 0.188,
    655: 0.196,
    660: 0.204,
    665: 0.213,
    670: 0.222,
    675: 0.231,
    680: 0.242,
    685: 0.251,
    690: 0.261,
    695: 0.271,
    700: 0.282,
    705: 0.294,
    710: 0.305,
    715: 0.318,
    720: 0.334,
    725: 0.354,
    730: 0.372,
    735: 0.392,
    740: 0.409,
    745: 0.420,
    750: 0.436,
    755: 0.450,
    760: 0.462,
    765: 0.465,
    770: 0.448,
    775: 0.432,
    780: 0.421,
}

sd_sample = colour.SpectralDistribution(data_sample, name="Sample")

message_box("Sample spectral distribution shape.")
print(sd_sample.shape)

print("\n")

message_box("Sample spectral distribution uniformity.")
print(sd_sample.is_uniform())

print("\n")

message_box(
    "Sample spectral distribution cloning:\n\n"
    '\t("Original Id", "Clone Id")\n'
    "\nCloning is a convenient way to get a copy of the spectral "
    "distribution, this an important feature because some operations happen "
    "in place."
)
sd_clone = sd_sample.copy()
print(id(sd_sample), id(sd_clone))

print("\n")

message_box("Sample spectral distribution arithmetical operations.")
message_box("Regular arithmetical operation: adding a constant.")
sd_clone_alternate = sd_clone + 10
print(sd_clone[380], sd_clone_alternate[380])

print("\n")

message_box("Regular arithmetical operation: adding an array.")
print((sd_clone + np.linspace(0, 1, len(sd_clone.wavelengths))).values)

print("\n")

message_box(
    "Regular arithmetical operation: adding a spectral " "distribution."
)
print((sd_clone + sd_clone).values)

print("\n")

message_box("In-place arithmetical operation: adding a constant.")
sd_clone += 10
print(sd_clone[380])

print("\n")

message_box("In-place arithmetical operation: adding an array.")
sd_clone += np.linspace(0, 1, len(sd_clone.wavelengths))
print(sd_clone.values)

print("\n")

message_box(
    "In-place arithmetical operation: adding a spectral " "distribution."
)
sd_clone += sd_clone
print(sd_clone.values)

print("\n")

message_box("Sample spectral distribution interpolation.")
sd_clone.interpolate(colour.SpectralShape(360, 780, 1))
print(sd_clone[666])

print("\n")

message_box("Sample spectral distribution extrapolation.")
sd_clone.extrapolate(colour.SpectralShape(340, 830, 1))
print(sd_clone[340], sd_clone[360])

print("\n")

message_box("Sample spectral distribution align.")
sd_clone.align(colour.SpectralShape(400, 700, 5))
print(sd_clone[400], sd_clone[700])

print("\n")

message_box("Constant value filled spectral distribution.")
print(colour.sd_constant(3.1415)[400])

print("\n")

message_box("Zeros filled spectral distribution.")
print(colour.sd_zeros()[400])

print("\n")

message_box("Ones filled spectral distribution.")
print(colour.sd_ones()[400])



"""
ITU-R BT.2100
=============

Defines the *ITU-R BT.2100* opto-electrical transfer functions (OETF),
opto-optical transfer functions (OOTF / OOCF) and electro-optical transfer
functions (EOTF) and their inverse:

-   :func:`colour.models.oetf_PQ_BT2100`
-   :func:`colour.models.oetf_inverse_PQ_BT2100`
-   :func:`colour.models.eotf_PQ_BT2100`
-   :func:`colour.models.eotf_inverse_PQ_BT2100`
-   :func:`colour.models.ootf_PQ_BT2100`
-   :func:`colour.models.ootf_inverse_PQ_BT2100`
-   :func:`colour.models.oetf_HLG_BT2100`
-   :func:`colour.models.oetf_inverse_HLG_BT2100`
-   :func:`colour.models.eotf_HLG_BT2100_1`
-   :func:`colour.models.eotf_HLG_BT2100_2`
-   :attr:`colour.models.BT2100_HLG_EOTF_METHODS`
-   :func:`colour.models.eotf_HLG_BT2100`
-   :func:`colour.models.eotf_inverse_HLG_BT2100_1`
-   :func:`colour.models.eotf_inverse_HLG_BT2100_2`
-   :attr:`colour.models.BT2100_HLG_EOTF_INVERSE_METHODS`
-   :func:`colour.models.eotf_inverse_HLG_BT2100`
-   :func:`colour.models.ootf_HLG_BT2100`
-   :func:`colour.models.ootf_inverse_HLG_BT2100`
-   :func:`colour.models.ootf_HLG_BT2100_1`
-   :func:`colour.models.ootf_HLG_BT2100_2`
-   :attr:`colour.models.BT2100_HLG_OOTF_METHODS`
-   :func:`colour.models.ootf_HLG_BT2100`
-   :func:`colour.models.ootf_inverse_HLG_BT2100_1`
-   :func:`colour.models.ootf_inverse_HLG_BT2100_2`
-   :attr:`colour.models.BT2100_HLG_OOTF_INVERSE_METHODS`
-   :func:`colour.models.ootf_inverse_HLG_BT2100`

References
----------
-   :cite:`Borer2017a` : Borer, T. (2017). Private Discussion with Mansencal,
    T. and Shaw, N.
-   :cite:`InternationalTelecommunicationUnion2017` : International
    Telecommunication Union. (2017). Recommendation ITU-R BT.2100-1 - Image
    parameter values for high dynamic range television for use in production
    and international programme exchange.
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2100-1-201706-I!!PDF-E.pdf
-   :cite:`InternationalTelecommunicationUnion2018` : International
    Telecommunication Union. (2018). Recommendation ITU-R BT.2100-2 - Image
    parameter values for high dynamic range television for use in production
    and international programme exchange.
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2100-2-201807-I!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.hints import (
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    NDArray,
    Optional,
    Union,
)
from colour.models.rgb.transfer_functions import (
    eotf_BT1886,
    eotf_ST2084,
    eotf_inverse_BT1886,
    oetf_ARIBSTDB67,
    oetf_BT709,
    eotf_inverse_ST2084,
    oetf_inverse_ARIBSTDB67,
    oetf_inverse_BT709,
)
from colour.models.rgb.transfer_functions.arib_std_b67 import (
    CONSTANTS_ARIBSTDB67,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    Structure,
    as_float,
    as_float_array,
    as_float_scalar,
    domain_range_scale,
    filter_kwargs,
    from_range_1,
    optional,
    to_domain_1,
    tsplit,
    tstack,
    usage_warning,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "oetf_PQ_BT2100",
    "oetf_inverse_PQ_BT2100",
    "eotf_PQ_BT2100",
    "eotf_inverse_PQ_BT2100",
    "ootf_PQ_BT2100",
    "ootf_inverse_PQ_BT2100",
    "WEIGHTS_BT2100_HLG",
    "CONSTANTS_BT2100_HLG",
    "gamma_function_HLG_BT2100",
    "oetf_HLG_BT2100",
    "oetf_inverse_HLG_BT2100",
    "black_level_lift_HLG_BT2100",
    "eotf_HLG_BT2100_1",
    "eotf_HLG_BT2100_2",
    "BT2100_HLG_EOTF_METHODS",
    "eotf_HLG_BT2100",
    "eotf_inverse_HLG_BT2100_1",
    "eotf_inverse_HLG_BT2100_2",
    "BT2100_HLG_EOTF_INVERSE_METHODS",
    "eotf_inverse_HLG_BT2100",
    "ootf_HLG_BT2100_1",
    "ootf_HLG_BT2100_2",
    "BT2100_HLG_OOTF_METHODS",
    "ootf_HLG_BT2100",
    "ootf_inverse_HLG_BT2100_1",
    "ootf_inverse_HLG_BT2100_2",
    "BT2100_HLG_OOTF_INVERSE_METHODS",
    "ootf_inverse_HLG_BT2100",
]


def oetf_PQ_BT2100(E: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* opto-electrical
    transfer function (OETF).

    The OETF maps relative scene linear light into the non-linear *PQ* signal
    value.

    Parameters
    ----------
    E
        :math:`E = {R_S, G_S, B_S; Y_S; or I_S}` is the signal determined by
        scene light and scaled by camera exposure.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` is the resulting non-linear signal (:math:`R'`, :math:`G'`,
        :math:`B'`).

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> oetf_PQ_BT2100(0.1)  # doctest: +ELLIPSIS
    0.7247698...
    """

    return eotf_inverse_ST2084(ootf_PQ_BT2100(E), 10000)


def oetf_inverse_PQ_BT2100(E_p: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* inverse
    opto-electrical transfer function (OETF).

    Parameters
    ----------
    E_p
        :math:`E'` is the resulting non-linear signal (:math:`R'`, :math:`G'`,
        :math:`B'`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E = {R_S, G_S, B_S; Y_S; or I_S}` is the signal determined by
        scene light and scaled by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> oetf_inverse_PQ_BT2100(0.724769816665726)  # doctest: +ELLIPSIS
    0.0999999...
    """

    return ootf_inverse_PQ_BT2100(eotf_ST2084(E_p, 10000))


def eotf_PQ_BT2100(E_p: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* electro-optical
    transfer function (EOTF).

    The EOTF maps the non-linear *PQ* signal into display light.

    Parameters
    ----------
    E_p
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *PQ* space [0, 1].

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> eotf_PQ_BT2100(0.724769816665726)  # doctest: +ELLIPSIS
    779.9883608...
    """

    return eotf_ST2084(E_p, 10000)


def eotf_inverse_PQ_BT2100(F_D: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* inverse
    electro-optical transfer function (EOTF).

    Parameters
    ----------
    F_D
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *PQ* space [0, 1].

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> eotf_inverse_PQ_BT2100(779.988360834085370)  # doctest: +ELLIPSIS
    0.7247698...
    """

    return eotf_inverse_ST2084(F_D, 10000)


def ootf_PQ_BT2100(E: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* opto-optical transfer
    function (OOTF / OOCF).

    The OOTF maps relative scene linear light to display linear light.

    Parameters
    ----------
    E
        :math:`E = {R_S, G_S, B_S; Y_S; or I_S}` is the signal determined by
        scene light and scaled by camera exposure.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`F_D` is the luminance of a displayed linear component
        (:math:`R_D`, :math:`G_D`, :math:`B_D`; :math:`Y_D`; or :math:`I_D`).

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> ootf_PQ_BT2100(0.1)  # doctest: +ELLIPSIS
    779.9883608...
    """

    E = as_float_array(E)

    with domain_range_scale("ignore"):
        return 100 * eotf_BT1886(oetf_BT709(59.5208 * E))


def ootf_inverse_PQ_BT2100(F_D: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference PQ* inverse opto-optical
    transfer function (OOTF / OOCF).

    Parameters
    ----------
    F_D
        :math:`F_D` is the luminance of a displayed linear component
        (:math:`R_D`, :math:`G_D`, :math:`B_D`; :math:`Y_D`; or :math:`I_D`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E = {R_S, G_S, B_S; Y_S; or I_S}` is the signal determined by
        scene light and scaled by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> ootf_inverse_PQ_BT2100(779.988360834115840)  # doctest: +ELLIPSIS
    0.1000000...
    """

    F_D = as_float_array(F_D)

    with domain_range_scale("ignore"):
        return oetf_inverse_BT709(eotf_inverse_BT1886(F_D / 100)) / 59.5208


WEIGHTS_BT2100_HLG: NDArray = np.array([0.2627, 0.6780, 0.0593])
"""Luminance weights for *Recommendation ITU-R BT.2100* *Reference HLG*."""

CONSTANTS_BT2100_HLG: Structure = Structure(
    a=CONSTANTS_ARIBSTDB67.a,
    b=1 - 4 * CONSTANTS_ARIBSTDB67.a,
    c=0.5 - CONSTANTS_ARIBSTDB67.a * np.log(4 * CONSTANTS_ARIBSTDB67.a),
)
"""
*Recommendation ITU-R BT.2100* *Reference HLG* constants expressed in their
analytical form in contrast to the *ARIB STD-B67 (Hybrid Log-Gamma)* numerical
reference.

References
----------
:cite:`InternationalTelecommunicationUnion2017`
"""


def gamma_function_HLG_BT2100(L_W: Floating = 1000) -> Floating:
    """
    Return the *Reference HLG* system gamma value for given display nominal
    peak luminance.

    Parameters
    ----------
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.

    Returns
    -------
    :class:`numpy.floating`
        *Reference HLG* system gamma value.

    Examples
    --------
    >>> gamma_function_HLG_BT2100()
    1.2
    >>> gamma_function_HLG_BT2100(2000)  # doctest: +ELLIPSIS
    1.3264325...
    >>> gamma_function_HLG_BT2100(4000)  # doctest: +ELLIPSIS
    1.4528651...
    """

    gamma = 1.2 + 0.42 * np.log10(L_W / 1000)

    return as_float_scalar(gamma)


def oetf_HLG_BT2100(
    E: FloatingOrArrayLike, constants: Structure = CONSTANTS_BT2100_HLG
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* opto-electrical
    transfer function (OETF).

    The OETF maps relative scene linear light into the non-linear *HLG* signal
    value.

    Parameters
    ----------
    E
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` is the resulting non-linear signal :math:`{R', G', B'}`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> oetf_HLG_BT2100(0.18 / 12)  # doctest: +ELLIPSIS
    0.2121320...
    """

    E = as_float_array(E)

    return oetf_ARIBSTDB67(12 * E, constants=constants)


def oetf_inverse_HLG_BT2100(
    E_p: FloatingOrArrayLike, constants: Structure = CONSTANTS_BT2100_HLG
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse
    opto-electrical transfer function (OETF).

    Parameters
    ----------
    E_p
        :math:`E'` is the resulting non-linear signal :math:`{R', G', B'}`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> oetf_inverse_HLG_BT2100(0.212132034355964)  # doctest: +ELLIPSIS
    0.0149999...
    """

    return oetf_inverse_ARIBSTDB67(E_p, constants=constants) / 12


def black_level_lift_HLG_BT2100(
    L_B: Floating = 0, L_W: Floating = 1000, gamma: Optional[Floating] = None
) -> Floating:
    """
    Return the *Reference HLG* black level lift :math:`\\beta` for given
    display luminance for black, nominal peak luminance and system gamma value.

    Parameters
    ----------
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.

    Returns
    -------
    :class:`numpy.floating`
        *Reference HLG* black level lift :math:`\\beta`.

    Examples
    --------
    >>> black_level_lift_HLG_BT2100()
    0.0
    >>> black_level_lift_HLG_BT2100(0.01)  # doctest: +ELLIPSIS
    0.0142964...
    >>> black_level_lift_HLG_BT2100(0.001, 2000)  # doctest: +ELLIPSIS
    0.0073009...
    >>> black_level_lift_HLG_BT2100(0.01, gamma=1.4)  # doctest: +ELLIPSIS
    0.0283691...
    """

    gamma = optional(gamma, gamma_function_HLG_BT2100(L_W))

    beta = np.sqrt(3 * spow((L_B / L_W), 1 / gamma))

    return as_float_scalar(beta)


def eotf_HLG_BT2100_1(
    E_p: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* electro-optical
    transfer function (EOTF) as given in *ITU-R BT.2100-1*.

    The EOTF maps the non-linear *HLG* signal into display light.

    Parameters
    ----------
    E_p
        :math:`E'` is the non-linear signal :math:`{R', G', B'}` as defined for
        the OETF.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> eotf_HLG_BT2100_1(0.212132034355964)  # doctest: +ELLIPSIS
    6.4760398...
    >>> eotf_HLG_BT2100_1(0.212132034355964, 0.01)  # doctest: +ELLIPSIS
    6.4859750...
    """

    return ootf_HLG_BT2100_1(
        oetf_inverse_ARIBSTDB67(E_p, constants=constants) / 12, L_B, L_W, gamma
    )


def eotf_HLG_BT2100_2(
    E_p: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* electro-optical
    transfer function (EOTF) as given in *ITU-R BT.2100-2* with the
    modified black level behaviour.

    The EOTF maps the non-linear *HLG* signal into display light.

    Parameters
    ----------
    E_p
        :math:`E'` is the non-linear signal :math:`{R', G', B'}` as defined for
        the *HLG Reference* OETF.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> eotf_HLG_BT2100_2(0.212132034355964)  # doctest: +ELLIPSIS
    6.4760398...
    >>> eotf_HLG_BT2100_2(0.212132034355964, 0.01)  # doctest: +ELLIPSIS
    7.3321975...
    """

    E_p = as_float_array(E_p)

    beta = black_level_lift_HLG_BT2100(L_B, L_W, gamma)

    return ootf_HLG_BT2100_2(
        oetf_inverse_ARIBSTDB67((1 - beta) * E_p + beta, constants=constants)
        / 12,
        L_W,
        gamma,
    )


BT2100_HLG_EOTF_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ITU-R BT.2100-1": eotf_HLG_BT2100_1,
        "ITU-R BT.2100-2": eotf_HLG_BT2100_2,
    }
)
BT2100_HLG_EOTF_METHODS.__doc__ = """
Supported *Recommendation ITU-R BT.2100* *Reference HLG* electro-optical
transfer function (EOTF).

References
----------
:cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
:cite:`InternationalTelecommunicationUnion2018`
"""


def eotf_HLG_BT2100(
    E_p: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
    method: Union[
        Literal["ITU-R BT.2100-1", "ITU-R BT.2100-2"], str
    ] = "ITU-R BT.2100-2",
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* electro-optical
    transfer function (EOTF).

    The EOTF maps the non-linear *HLG* signal into display light.

    Parameters
    ----------
    E_p
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *HLG* space.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
    :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> eotf_HLG_BT2100(0.212132034355964)  # doctest: +ELLIPSIS
    6.4760398...
    >>> eotf_HLG_BT2100(0.212132034355964, method='ITU-R BT.2100-1')
    ... # doctest: +ELLIPSIS
    6.4760398...
    >>> eotf_HLG_BT2100(0.212132034355964, 0.01)
    ... # doctest: +ELLIPSIS
    7.3321975...
    """

    method = validate_method(method, BT2100_HLG_EOTF_METHODS)

    return BT2100_HLG_EOTF_METHODS[method](E_p, L_B, L_W, gamma, constants)


def eotf_inverse_HLG_BT2100_1(
    F_D: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse
    electro-optical transfer function (EOTF) as given in
    *ITU-R BT.2100-1*.

    Parameters
    ----------
    F_D
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *HLG* space.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> eotf_inverse_HLG_BT2100_1(6.476039825649814)  # doctest: +ELLIPSIS
    0.2121320...
    >>> eotf_inverse_HLG_BT2100_1(6.485975065251558, 0.01)
    ... # doctest: +ELLIPSIS
    0.2121320...
    """

    return oetf_ARIBSTDB67(
        ootf_inverse_HLG_BT2100_1(F_D, L_B, L_W, gamma) * 12,
        constants=constants,
    )


def eotf_inverse_HLG_BT2100_2(
    F_D: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse
    electro-optical transfer function (EOTF) as given in
    *ITU-R BT.2100-2* with the modified black level behaviour.

    Parameters
    ----------
    F_D
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *HLG* space.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> eotf_inverse_HLG_BT2100_2(6.476039825649814)  # doctest: +ELLIPSIS
    0.2121320...
    >>> eotf_inverse_HLG_BT2100_2(7.332197528353875, 0.01)
    ... # doctest: +ELLIPSIS
    0.2121320...
    """

    beta = black_level_lift_HLG_BT2100(L_B, L_W, gamma)

    return (
        oetf_ARIBSTDB67(
            ootf_inverse_HLG_BT2100_2(F_D, L_W, gamma) * 12,
            constants=constants,
        )
        - beta
    ) / (1 - beta)


BT2100_HLG_EOTF_INVERSE_METHODS: CaseInsensitiveMapping = (
    CaseInsensitiveMapping(
        {
            "ITU-R BT.2100-1": eotf_inverse_HLG_BT2100_1,
            "ITU-R BT.2100-2": eotf_inverse_HLG_BT2100_2,
        }
    )
)
BT2100_HLG_EOTF_INVERSE_METHODS.__doc__ = """
Supported *Recommendation ITU-R BT.2100* *Reference HLG* inverse
electro-optical transfer function (EOTF).

References
----------
:cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
:cite:`InternationalTelecommunicationUnion2018`
"""


def eotf_inverse_HLG_BT2100(
    F_D: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    constants: Structure = CONSTANTS_BT2100_HLG,
    method: Union[
        Literal["ITU-R BT.2100-1", "ITU-R BT.2100-2"], str
    ] = "ITU-R BT.2100-2",
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse
    electro-optical transfer function (EOTF).

    Parameters
    ----------
    F_D
        Luminance :math:`F_D` of a displayed linear component
        :math:`{R_D, G_D, B_D}` or :math:`Y_D` or :math:`I_D`, in
        :math:`cd/m^2`.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    constants
        *Recommendation ITU-R BT.2100* *Reference HLG* constants.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E'` denotes a non-linear colour value :math:`{R', G', B'}` or
        :math:`{L', M', S'}` in *HLG* space.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E_p``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
    :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> eotf_inverse_HLG_BT2100(6.476039825649814)  # doctest: +ELLIPSIS
    0.2121320...
    >>> eotf_inverse_HLG_BT2100(6.476039825649814, method='ITU-R BT.2100-1')
    ... # doctest: +ELLIPSIS
    0.2121320...
    >>> eotf_inverse_HLG_BT2100(7.332197528353875, 0.01)  # doctest: +ELLIPSIS
    0.2121320...
    """

    method = validate_method(method, BT2100_HLG_EOTF_INVERSE_METHODS)

    return BT2100_HLG_EOTF_INVERSE_METHODS[method](
        F_D, L_B, L_W, gamma, constants
    )


def ootf_HLG_BT2100_1(
    E: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* opto-optical
    transfer function (OOTF / OOCF) as given in *ITU-R BT.2100-1*.

    The OOTF maps relative scene linear light to display linear light.

    Parameters
    ----------
    E
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> ootf_HLG_BT2100_1(0.1)  # doctest: +ELLIPSIS
    63.0957344...
    >>> ootf_HLG_BT2100_1(0.1, 0.01)
    ... # doctest: +ELLIPSIS
    63.1051034...
    """

    E = as_float_array(np.atleast_1d(to_domain_1(E)))

    if E.shape[-1] != 3:
        usage_warning(
            '"Recommendation ITU-R BT.2100" "Reference HLG OOTF" uses '
            "RGB Luminance in computations and expects a vector input, thus "
            "the given input array will be stacked to compose a vector for "
            "internal computations but a single component will be output."
        )
        R_S = G_S = B_S = E
    else:
        R_S, G_S, B_S = tsplit(E)

    alpha = L_W - L_B
    beta = L_B

    Y_S = np.sum(WEIGHTS_BT2100_HLG * tstack([R_S, G_S, B_S]), axis=-1)

    gamma = optional(gamma, gamma_function_HLG_BT2100(L_W))

    R_D = alpha * R_S * np.abs(Y_S) ** (gamma - 1) + beta
    G_D = alpha * G_S * np.abs(Y_S) ** (gamma - 1) + beta
    B_D = alpha * B_S * np.abs(Y_S) ** (gamma - 1) + beta

    if E.shape[-1] != 3:
        return as_float(from_range_1(R_D))
    else:
        RGB_D = tstack([R_D, G_D, B_D])

        return from_range_1(RGB_D)


def ootf_HLG_BT2100_2(
    E: FloatingOrArrayLike,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* opto-optical
    transfer function (OOTF / OOCF) as given in *ITU-R BT.2100-2*.

    The OOTF maps relative scene linear light to display linear light.

    Parameters
    ----------
    E
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> ootf_HLG_BT2100_2(0.1)  # doctest: +ELLIPSIS
    63.0957344...
    """

    E = as_float_array(np.atleast_1d(to_domain_1(E)))

    if E.shape[-1] != 3:
        usage_warning(
            '"Recommendation ITU-R BT.2100" "Reference HLG OOTF" uses '
            "RGB Luminance in computations and expects a vector input, thus "
            "the given input array will be stacked to compose a vector for "
            "internal computations but a single component will be output."
        )
        R_S = G_S = B_S = E
    else:
        R_S, G_S, B_S = tsplit(E)

    alpha = L_W

    Y_S = np.sum(WEIGHTS_BT2100_HLG * tstack([R_S, G_S, B_S]), axis=-1)

    gamma = optional(gamma, gamma_function_HLG_BT2100(L_W))

    R_D = alpha * R_S * np.abs(Y_S) ** (gamma - 1)
    G_D = alpha * G_S * np.abs(Y_S) ** (gamma - 1)
    B_D = alpha * B_S * np.abs(Y_S) ** (gamma - 1)

    if E.shape[-1] != 3:
        return as_float(from_range_1(R_D))
    else:
        RGB_D = tstack([R_D, G_D, B_D])

        return from_range_1(RGB_D)


BT2100_HLG_OOTF_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ITU-R BT.2100-1": ootf_HLG_BT2100_1,
        "ITU-R BT.2100-2": ootf_HLG_BT2100_2,
    }
)
BT2100_HLG_OOTF_METHODS.__doc__ = """
Supported *Recommendation ITU-R BT.2100* *Reference HLG* opto-optical transfer
function (OOTF / OOCF).

References
----------
:cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
:cite:`InternationalTelecommunicationUnion2018`
"""


def ootf_HLG_BT2100(
    E: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    method: Union[
        Literal["ITU-R BT.2100-1", "ITU-R BT.2100-2"], str
    ] = "ITU-R BT.2100-2",
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* opto-optical
    transfer function (OOTF / OOCF).

    The OOTF maps relative scene linear light to display linear light.

    Parameters
    ----------
    E
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> ootf_HLG_BT2100(0.1)  # doctest: +ELLIPSIS
    63.0957344...
    >>> ootf_HLG_BT2100(0.1, 0.01, method='ITU-R BT.2100-1')
    ... # doctest: +ELLIPSIS
    63.1051034...
    """

    method = validate_method(method, BT2100_HLG_OOTF_METHODS)

    function = BT2100_HLG_OOTF_METHODS[method]

    return function(
        E,
        **filter_kwargs(function, **{"L_B": L_B, "L_W": L_W, "gamma": gamma}),
    )


def ootf_inverse_HLG_BT2100_1(
    F_D: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse opto-optical
    transfer function (OOTF / OOCF) as given in *ITU-R BT.2100-1*.

    Parameters
    ----------
    F_D
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`

    Examples
    --------
    >>> ootf_inverse_HLG_BT2100_1(63.095734448019336)  # doctest: +ELLIPSIS
    0.1000000...
    >>> ootf_inverse_HLG_BT2100_1(63.105103490674857, 0.01)
    ... # doctest: +ELLIPSIS
    0.0999999...
    """

    F_D = as_float_array(np.atleast_1d(to_domain_1(F_D)))

    if F_D.shape[-1] != 3:
        usage_warning(
            '"Recommendation ITU-R BT.2100" "Reference HLG OOTF" uses '
            "RGB Luminance in computations and expects a vector input, thus "
            "the given input array will be stacked to compose a vector for "
            "internal computations but a single component will be output."
        )
        R_D = G_D = B_D = F_D
    else:
        R_D, G_D, B_D = tsplit(F_D)

    Y_D = np.sum(WEIGHTS_BT2100_HLG * tstack([R_D, G_D, B_D]), axis=-1)

    alpha = L_W - L_B
    beta = L_B

    gamma = optional(gamma, gamma_function_HLG_BT2100(L_W))

    Y_D_beta = np.abs((Y_D - beta) / alpha) ** ((1 - gamma) / gamma)

    R_S = np.where(
        Y_D == beta,
        0.0,
        Y_D_beta * (R_D - beta) / alpha,
    )
    G_S = np.where(
        Y_D == beta,
        0.0,
        Y_D_beta * (G_D - beta) / alpha,
    )
    B_S = np.where(
        Y_D == beta,
        0.0,
        Y_D_beta * (B_D - beta) / alpha,
    )

    if F_D.shape[-1] != 3:
        return as_float(from_range_1(R_S))
    else:
        RGB_S = tstack([R_S, G_S, B_S])

        return from_range_1(RGB_S)


def ootf_inverse_HLG_BT2100_2(
    F_D: FloatingOrArrayLike,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse opto-optical
    transfer function (OOTF / OOCF) as given in *ITU-R BT.2100-2*.

    Parameters
    ----------
    F_D
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> ootf_inverse_HLG_BT2100_2(63.095734448019336)  # doctest: +ELLIPSIS
    0.1000000...
    """

    F_D = as_float_array(np.atleast_1d(to_domain_1(F_D)))

    if F_D.shape[-1] != 3:
        usage_warning(
            '"Recommendation ITU-R BT.2100" "Reference HLG OOTF" uses '
            "RGB Luminance in computations and expects a vector input, thus "
            "the given input array will be stacked to compose a vector for "
            "internal computations but a single component will be output."
        )
        R_D = G_D = B_D = F_D
    else:
        R_D, G_D, B_D = tsplit(F_D)

    Y_D = np.sum(WEIGHTS_BT2100_HLG * tstack([R_D, G_D, B_D]), axis=-1)

    alpha = L_W

    gamma = optional(gamma, gamma_function_HLG_BT2100(L_W))

    Y_D_alpha = np.abs(Y_D / alpha) ** ((1 - gamma) / gamma)

    R_S = np.where(
        Y_D == 0,
        0.0,
        Y_D_alpha * R_D / alpha,
    )
    G_S = np.where(
        Y_D == 0,
        0.0,
        Y_D_alpha * G_D / alpha,
    )
    B_S = np.where(
        Y_D == 0,
        0.0,
        Y_D_alpha * B_D / alpha,
    )

    if F_D.shape[-1] != 3:
        return as_float(from_range_1(R_S))
    else:
        RGB_S = tstack([R_S, G_S, B_S])

        return from_range_1(RGB_S)


BT2100_HLG_OOTF_INVERSE_METHODS: CaseInsensitiveMapping = (
    CaseInsensitiveMapping(
        {
            "ITU-R BT.2100-1": ootf_inverse_HLG_BT2100_1,
            "ITU-R BT.2100-2": ootf_inverse_HLG_BT2100_2,
        }
    )
)
BT2100_HLG_OOTF_INVERSE_METHODS.__doc__ = """
Supported *Recommendation ITU-R BT.2100* *Reference HLG* inverse opto-optical
transfer function (OOTF / OOCF).

References
----------
:cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
:cite:`InternationalTelecommunicationUnion2018`
"""


def ootf_inverse_HLG_BT2100(
    F_D: FloatingOrArrayLike,
    L_B: Floating = 0,
    L_W: Floating = 1000,
    gamma: Optional[Floating] = None,
    method: Union[
        Literal["ITU-R BT.2100-1", "ITU-R BT.2100-2"], str
    ] = "ITU-R BT.2100-2",
) -> FloatingOrNDArray:
    """
    Define *Recommendation ITU-R BT.2100* *Reference HLG* inverse opto-optical
    transfer function (OOTF / OOCF).

    Parameters
    ----------
    F_D
        :math:`F_D` is the luminance of a displayed linear component
        :math:`{R_D, G_D, or B_D}`, in :math:`cd/m^2`.
    L_B
        :math:`L_B` is the display luminance for black in :math:`cd/m^2`.
    L_W
        :math:`L_W` is nominal peak luminance of the display in :math:`cd/m^2`
        for achromatic pixels.
    gamma
        System gamma value, 1.2 at the nominal display peak luminance of
        :math:`1000 cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`E` is the signal for each colour component
        :math:`{R_S, G_S, B_S}` proportional to scene linear light and scaled
        by camera exposure.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``F_D``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``E``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Borer2017a`, :cite:`InternationalTelecommunicationUnion2017`,
    :cite:`InternationalTelecommunicationUnion2018`

    Examples
    --------
    >>> ootf_inverse_HLG_BT2100(63.095734448019336)  # doctest: +ELLIPSIS
    0.1000000...
    >>> ootf_inverse_HLG_BT2100(
    ...     63.105103490674857, 0.01, method='ITU-R BT.2100-1')
    ... # doctest: +ELLIPSIS
    0.0999999...
    """

    method = validate_method(method, BT2100_HLG_OOTF_INVERSE_METHODS)

    function = BT2100_HLG_OOTF_INVERSE_METHODS[method]

    return function(
        F_D,
        **filter_kwargs(function, **{"L_B": L_B, "L_W": L_W, "gamma": gamma}),
    )



"""Defines the unit tests for the :mod:`colour.graph.conversion` module."""

import numpy as np
import unittest

from colour.characterisation import SDS_COLOURCHECKERS
from colour.colorimetry import CCS_ILLUMINANTS, SDS_ILLUMINANTS
from colour.models import COLOURSPACE_MODELS, RGB_COLOURSPACE_ACES2065_1
from colour.graph import describe_conversion_path, convert

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestDescribeConversionPath",
    "TestConvert",
]


class TestDescribeConversionPath(unittest.TestCase):
    """
    Define :func:`colour.graph.conversion.describe_conversion_path` definition
    unit tests methods.
    """

    def test_describe_conversion_path(self):
        """
        Test :func:`colour.graph.conversion.describe_conversion_path`
        definition.
        """

        describe_conversion_path("Spectral Distribution", "sRGB")

        describe_conversion_path("Spectral Distribution", "sRGB", mode="Long")

        describe_conversion_path(
            "Spectral Distribution",
            "sRGB",
            mode="Extended",
            sd_to_XYZ={
                "illuminant": SDS_ILLUMINANTS["FL2"],
                "return": np.array([0.47924575, 0.31676968, 0.17362725]),
            },
        )


class TestConvert(unittest.TestCase):
    """
    Define :func:`colour.graph.conversion.convert` definition unit tests
    methods.
    """

    def test_convert(self):
        """Test :func:`colour.graph.conversion.convert` definition."""

        RGB_a = convert(
            SDS_COLOURCHECKERS["ColorChecker N Ohta"]["dark skin"],
            "Spectral Distribution",
            "sRGB",
        )
        np.testing.assert_almost_equal(
            RGB_a, np.array([0.45675795, 0.30986982, 0.24861924]), decimal=7
        )

        Jpapbp = convert(RGB_a, "Output-Referred RGB", "CAM16UCS")
        np.testing.assert_almost_equal(
            Jpapbp, np.array([0.39994810, 0.09206557, 0.08127526]), decimal=7
        )

        RGB_b = convert(
            Jpapbp, "CAM16UCS", "sRGB", verbose={"mode": "Extended"}
        )
        # NOTE: The "CIE XYZ" tristimulus values to "sRGB" matrix is given
        # rounded at 4 decimals as per "IEC 61966-2-1:1999" and thus preventing
        # exact roundtrip.
        np.testing.assert_allclose(RGB_a, RGB_b, rtol=1e-5, atol=1e-5)

        np.testing.assert_almost_equal(
            convert("#808080", "Hexadecimal", "Scene-Referred RGB"),
            np.array([0.21586050, 0.21586050, 0.21586050]),
            decimal=7,
        )

        self.assertAlmostEqual(
            convert("#808080", "Hexadecimal", "RGB Luminance"),
            0.21586050,
            places=7,
        )

        np.testing.assert_almost_equal(
            convert(
                convert(
                    np.array([0.5, 0.5, 0.5]),
                    "Output-Referred RGB",
                    "Scene-Referred RGB",
                ),
                "RGB",
                "YCbCr",
            ),
            np.array([0.49215686, 0.50196078, 0.50196078]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            convert(
                RGB_a,
                "RGB",
                "Scene-Referred RGB",
                RGB_to_RGB={"output_colourspace": RGB_COLOURSPACE_ACES2065_1},
            ),
            np.array([0.36364180, 0.31715308, 0.25888531]),
            decimal=7,
        )

        # Consistency check to verify that all the colour models are properly
        # named in the graph:
        for model in COLOURSPACE_MODELS:
            convert(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                "CIE XYZ",
                model,
            )

    def test_convert_direct_keyword_argument_passing(self):
        """
        Test :func:`colour.graph.conversion.convert` definition behaviour when
        direct keyword arguments are passed.
        """

        a = np.array([0.20654008, 0.12197225, 0.05136952])
        illuminant = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D50"
        ]
        np.testing.assert_almost_equal(
            convert(
                a, "CIE XYZ", "CIE xyY", XYZ_to_xyY={"illuminant": illuminant}
            ),
            convert(a, "CIE XYZ", "CIE xyY", illuminant=illuminant),
            decimal=7,
        )

        # Illuminant "ndarray" is converted to tuple here so that it can
        # be hashed by the "sd_to_XYZ" definition, this should never occur
        # in practical application.
        self.assertRaises(
            AttributeError,
            lambda: convert(
                SDS_COLOURCHECKERS["ColorChecker N Ohta"]["dark skin"],
                "Spectral Distribution",
                "sRGB",
                illuminant=tuple(illuminant),
            ),
        )


if __name__ == "__main__":
    unittest.main()



"""Showcases *RLAB* colour appearance model computations."""

import numpy as np

import colour
from colour.appearance.rlab import CAM_ReferenceSpecification_RLAB
from colour.utilities import message_box

message_box('"RLAB" Colour Appearance Model Computations')

XYZ = np.array([19.01, 20.00, 21.78])
XYZ_n = np.array([109.85, 100, 35.58])
Y_n = 31.83
sigma = colour.VIEWING_CONDITIONS_RLAB["Average"]
D = colour.appearance.D_FACTOR_RLAB["Hard Copy Images"]
message_box(
    f'Converting to the "RLAB" colour appearance model specification using '
    f"given parameters:\n\n"
    f"\tXYZ: {XYZ}\n"
    f"\tXYZ_n: {XYZ_n}\n"
    f"\tY_n: {Y_n}\n"
    f"\tsigma: {sigma}\n"
    f"\tD: {D}"
)
specification = colour.XYZ_to_RLAB(XYZ, XYZ_n, Y_n, sigma, D)
print(specification)

print("\n")

message_box(
    'Broadcasting the current output "RLAB" colour appearance '
    "model specification to the reference specification.\n"
    "The intent of this reference specification is to provide names "
    'as closest as possible to the "Mark D. Fairchild" reference.\n'
    "The current output specification is meant to be consistent with "
    "the other colour appearance model specification by using same "
    "argument names for consistency wherever possible."
)

print(CAM_ReferenceSpecification_RLAB(*specification.values))



"""Defines the unit tests for the :mod:`colour.models.cam16_ucs` module."""

import unittest

from colour.models.tests.test_cam02_ucs import (
    TestJMh_CIECAM02_to_UCS_Luo2006,
    TestUCS_Luo2006_to_JMh_CIECAM02,
    TestXYZ_to_UCS_Luo2006,
    TestUCS_Luo2006_to_XYZ,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestJMh_CAM16_to_UCS_Li2017",
    "TestUCS_Li2017_to_JMh_CAM16",
    "TestXYZ_to_UCS_Li2017",
    "TestUCS_Li2017_to_XYZ",
]


class TestJMh_CAM16_to_UCS_Li2017(TestJMh_CIECAM02_to_UCS_Luo2006):
    """
    Define :func:`colour.models.cam16_ucs.JMh_CAM16_to_UCS_Li2017`
    definition unit tests methods.

    Notes
    -----
    -   :func:`colour.models.cam16_ucs.JMh_CAM16_to_UCS_Li2017` is a wrapper
        of :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006` and thus
        currently adopts the same unittests.
    """


class TestUCS_Li2017_to_JMh_CAM16(TestUCS_Luo2006_to_JMh_CIECAM02):
    """
    Define :func:`colour.models.cam16_ucs.UCS_Li2017_to_JMh_CAM16`
    definition unit tests methods.

    Notes
    -----
    -   :func:`colour.models.cam16_ucs.UCS_Li2017_to_JMh_CAM16` is a wrapper
        of :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02` and thus
        currently adopts the same unittests.
    """


class TestXYZ_to_UCS_Li2017(TestXYZ_to_UCS_Luo2006):
    """
    Define :func:`colour.models.cam16_ucs.XYZ_to_UCS_Li2017`
    definition unit tests methods.
    """

    pass


class TestUCS_Li2017_to_XYZ(TestUCS_Luo2006_to_XYZ):
    """
    Define :func:`colour.models.cam16_ucs.UCS_Li2017_to_XYZ`
    definition unit tests methods.
    """

    pass


if __name__ == "__main__":
    unittest.main()



"""
McCamy (1992) Correlated Colour Temperature
===========================================

Defines the *McCamy (1992)* correlated colour temperature :math:`T_{cp}`
computations objects:

-   :func:`colour.temperature.xy_to_CCT_McCamy1992`: Correlated colour
    temperature :math:`T_{cp}` computation of given *CIE xy* chromaticity
    coordinates using *McCamy (1992)* method.
-   :func:`colour.temperature.xy_to_CCT_McCamy1992`: *CIE xy* chromaticity
    coordinates computation of given correlated colour temperature
    :math:`T_{cp}` using *McCamy (1992)* method.

References
----------
-   :cite:`Wikipedia2001` : Wikipedia. (2001). Approximation. Retrieved June
    28, 2014, from http://en.wikipedia.org/wiki/Color_temperature#Approximation
"""

from __future__ import annotations

import numpy as np
from scipy.optimize import minimize

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import (
    ArrayLike,
    Dict,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Optional,
)
from colour.utilities import as_float_array, as_float, tsplit, usage_warning

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "xy_to_CCT_McCamy1992",
    "CCT_to_xy_McCamy1992",
]


def xy_to_CCT_McCamy1992(xy: ArrayLike) -> FloatingOrNDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` from given
    *CIE xy* chromaticity coordinates using *McCamy (1992)* method.

    Parameters
    ----------
    xy
        *CIE xy* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`.

    References
    ----------
    :cite:`Wikipedia2001`

    Examples
    --------
    >>> import numpy as np
    >>> xy = np.array([0.31270, 0.32900])
    >>> xy_to_CCT_McCamy1992(xy)  # doctest: +ELLIPSIS
    6505.0805913...
    """

    x, y = tsplit(xy)

    n = (x - 0.3320) / (y - 0.1858)
    CCT = -449 * n**3 + 3525 * n**2 - 6823.3 * n + 5520.33

    return as_float(CCT)


def CCT_to_xy_McCamy1992(
    CCT: FloatingOrArrayLike, optimisation_kwargs: Optional[Dict] = None
) -> NDArray:
    """
    Return the *CIE xy* chromaticity coordinates from given correlated colour
    temperature :math:`T_{cp}` using *McCamy (1992)* method.

    Parameters
    ----------
    CCT
        Correlated colour temperature :math:`T_{cp}`.
    optimisation_kwargs
        Parameters for :func:`scipy.optimize.minimize` definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE xy* chromaticity coordinates.

    Warnings
    --------
    *McCamy (1992)* method for computing *CIE xy* chromaticity coordinates
    from given correlated colour temperature is not a bijective function and
    might produce unexpected results. It is given for consistency with other
    correlated colour temperature computation methods but should be avoided
    for practical applications. The current implementation relies on
    optimization using :func:`scipy.optimize.minimize` definition and thus has
    reduced precision and poor performance.

    References
    ----------
    :cite:`Wikipedia2001`

    Examples
    --------
    >>> CCT_to_xy_McCamy1992(6505.0805913074782)  # doctest: +ELLIPSIS
    array([ 0.3127...,  0.329...])
    """

    usage_warning(
        '"McCamy (1992)" method for computing "CIE xy" '
        "chromaticity coordinates from given correlated colour "
        "temperature is not a bijective function and might produce "
        "unexpected results. It is given for consistency with other "
        "correlated colour temperature computation methods but "
        "should be avoided for practical applications."
    )

    CCT = as_float_array(CCT)
    shape = list(CCT.shape)
    CCT = np.atleast_1d(CCT.reshape([-1, 1]))

    def objective_function(
        xy: ArrayLike, CCT: FloatingOrArrayLike
    ) -> FloatingOrNDArray:
        """Objective function."""

        objective = np.linalg.norm(
            xy_to_CCT_McCamy1992(xy) - as_float_array(CCT)
        )

        return as_float(objective)

    optimisation_settings = {
        "method": "Nelder-Mead",
        "options": {
            "fatol": 1e-10,
        },
    }
    if optimisation_kwargs is not None:
        optimisation_settings.update(optimisation_kwargs)

    xy = as_float_array(
        [
            minimize(
                objective_function,
                x0=CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
                    "D65"
                ],
                args=(CCT_i,),
                **optimisation_settings,
            ).x
            for CCT_i in as_float_array(CCT)
        ]
    )

    return np.reshape(xy, (shape + [2]))



"""
References
----------
-   :cite:`AdobeSystems2013b` : Adobe Systems. (2013). Cube LUT Specification.
    https://drive.google.com/open?id=143Eh08ZYncCAMwJ1q4gWxVOqR_OSWYvs
-   :cite:`Chamberlain2015` : Chamberlain, P. (2015). LUT documentation (to
    create from another program). Retrieved August 23, 2018, from
    https://forum.blackmagicdesign.com/viewtopic.php?f=21&t=40284#p232952
-   :cite:`RisingSunResearch` : Rising Sun Research. (n.d.). cineSpace LUT
    Library. Retrieved November 30, 2018, from
    https://sourceforge.net/projects/cinespacelutlib/
"""

from __future__ import annotations

import os

from colour.hints import Any, Boolean, Integer, Literal, Optional, Union, cast
from colour.utilities import (
    CaseInsensitiveMapping,
    filter_kwargs,
    optional,
    validate_method,
)

from .lut import LUT1D, LUT3x1D, LUT3D, LUT_to_LUT
from .operator import AbstractLUTSequenceOperator, LUTOperatorMatrix
from .sequence import LUTSequence
from .iridas_cube import read_LUT_IridasCube, write_LUT_IridasCube
from .resolve_cube import read_LUT_ResolveCube, write_LUT_ResolveCube
from .sony_spi1d import read_LUT_SonySPI1D, write_LUT_SonySPI1D
from .sony_spi3d import read_LUT_SonySPI3D, write_LUT_SonySPI3D
from .sony_spimtx import read_LUT_SonySPImtx, write_LUT_SonySPImtx
from .cinespace_csp import read_LUT_Cinespace, write_LUT_Cinespace

__all__ = [
    "LUT1D",
    "LUT3x1D",
    "LUT3D",
    "LUT_to_LUT",
]
__all__ += [
    "AbstractLUTSequenceOperator",
    "LUTOperatorMatrix",
]
__all__ += [
    "LUTSequence",
]
__all__ += [
    "read_LUT_IridasCube",
    "write_LUT_IridasCube",
]
__all__ += [
    "read_LUT_ResolveCube",
    "write_LUT_ResolveCube",
]
__all__ += [
    "read_LUT_SonySPI1D",
    "write_LUT_SonySPI1D",
]
__all__ += [
    "read_LUT_SonySPI3D",
    "write_LUT_SonySPI3D",
]
__all__ += [
    "read_LUT_SonySPImtx",
    "write_LUT_SonySPImtx",
]
__all__ += [
    "read_LUT_Cinespace",
    "write_LUT_Cinespace",
]

MAPPING_EXTENSION_TO_LUT_FORMAT: CaseInsensitiveMapping = (
    CaseInsensitiveMapping(
        {
            ".cube": "Iridas Cube",
            ".spi1d": "Sony SPI1D",
            ".spi3d": "Sony SPI3D",
            ".spimtx": "Sony SPImtx",
            ".csp": "Cinespace",
        }
    )
)
"""Extension to *LUT* format."""

LUT_READ_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Cinespace": read_LUT_Cinespace,
        "Iridas Cube": read_LUT_IridasCube,
        "Resolve Cube": read_LUT_ResolveCube,
        "Sony SPI1D": read_LUT_SonySPI1D,
        "Sony SPI3D": read_LUT_SonySPI3D,
        "Sony SPImtx": read_LUT_SonySPImtx,
    }
)
LUT_READ_METHODS.__doc__ = """
Supported *LUT* reading methods.

References
----------
:cite:`AdobeSystems2013b`, :cite:`Chamberlain2015`
"""


def read_LUT(
    path: str,
    method: Optional[
        Union[
            Literal[
                "Cinespace",
                "Iridas Cube",
                "Resolve Cube",
                "Sony SPI1D",
                "Sony SPI3D",
                "Sony SPImtx",
            ],
            str,
        ]
    ] = None,
    **kwargs: Any,
) -> Union[LUT1D, LUT3x1D, LUT3D, LUTSequence, LUTOperatorMatrix]:
    """
    Read given *LUT* file using given method.

    Parameters
    ----------
    path
        *LUT* path.
    method
        Reading method, if *None*, the method will be auto-detected according
        to extension.

    Returns
    -------
    :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or :class:`colour.LUT3D` \
or :class:`colour.LUTSequence` or :class:`colour.LUTOperatorMatrix`
        :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or
        :class:`colour.LUT3D` or :class:`colour.LUTSequence` or
        :class:`colour.LUTOperatorMatrix` class instance.

    References
    ----------
    :cite:`AdobeSystems2013b`, :cite:`Chamberlain2015`,
    :cite:`RisingSunResearch`

    Examples
    --------
    Reading a 3x1D *Iridas* *.cube* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'iridas_cube',
    ...     'ACES_Proxy_10_to_ACES.cube')
    >>> print(read_LUT(path))
    LUT3x1D - ACES Proxy 10 to ACES
    -------------------------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (32, 3)

    Reading a 1D *Sony* *.spi1d* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'sony_spi1d',
    ...     'eotf_sRGB_1D.spi1d')
    >>> print(read_LUT(path))
    LUT1D - eotf sRGB 1D
    --------------------
    <BLANKLINE>
    Dimensions : 1
    Domain     : [-0.1  1.5]
    Size       : (16,)
    Comment 01 : Generated by "Colour 0.3.11".
    Comment 02 : "colour.models.eotf_sRGB".

    Reading a 3D *Sony* *.spi3d* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'sony_spi3d',
    ...     'Colour_Correct.spi3d')
    >>> print(read_LUT(path))
    LUT3D - Colour Correct
    ----------------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (4, 4, 4, 3)
    Comment 01 : Adapted from a LUT generated by Foundry::LUT.

    Reading a *Sony* *.spimtx* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'sony_spimtx',
    ...     'dt.spimtx')
    >>> print(read_LUT(path))
    LUTOperatorMatrix - dt
    ----------------------
    <BLANKLINE>
    Matrix     : [[ 0.864274  0.        0.        0.      ]
                  [ 0.        0.864274  0.        0.      ]
                  [ 0.        0.        0.864274  0.      ]
                  [ 0.        0.        0.        1.      ]]
    Offset     : [ 0.  0.  0.  0.]
    """

    method = cast(
        str,
        optional(
            method, MAPPING_EXTENSION_TO_LUT_FORMAT[os.path.splitext(path)[-1]]
        ),
    )

    method = validate_method(method, LUT_READ_METHODS)

    function = LUT_READ_METHODS[method]

    try:
        return function(path, **filter_kwargs(function, **kwargs))
    except ValueError as error:
        # Case where a "Resolve Cube" with "LUT3x1D" shaper was read as an
        # "Iridas Cube" "LUT".
        if method == "iridas cube":
            function = LUT_READ_METHODS["Resolve Cube"]
            return function(path, **filter_kwargs(function, **kwargs))
        else:
            raise error


LUT_WRITE_METHODS = CaseInsensitiveMapping(
    {
        "Cinespace": write_LUT_Cinespace,
        "Iridas Cube": write_LUT_IridasCube,
        "Resolve Cube": write_LUT_ResolveCube,
        "Sony SPI1D": write_LUT_SonySPI1D,
        "Sony SPI3D": write_LUT_SonySPI3D,
        "Sony SPImtx": write_LUT_SonySPImtx,
    }
)
LUT_WRITE_METHODS.__doc__ = """
Supported *LUT* reading methods.

References
----------
:cite:`AdobeSystems2013b`, :cite:`Chamberlain2015`
"""


def write_LUT(
    LUT: Union[LUT1D, LUT3x1D, LUT3D, LUTSequence, LUTOperatorMatrix],
    path: str,
    decimals: Integer = 7,
    method: Optional[
        Union[
            Literal[
                "Cinespace",
                "Iridas Cube",
                "Resolve Cube",
                "Sony SPI1D",
                "Sony SPI3D",
                "Sony SPImtx",
            ],
            str,
        ]
    ] = None,
    **kwargs: Any,
) -> Boolean:
    """
    Write given *LUT* to given file using given method.

    Parameters
    ----------
    LUT
        :class:`colour.LUT1D` or :class:`colour.LUT3x1D` or
        :class:`colour.LUT3D` or :class:`colour.LUTSequence` or
        :class:`colour.LUTOperatorMatrix` class instance to write at given
        path.
    path
        *LUT* path.
    decimals
        Formatting decimals.
    method
        Writing method, if *None*, the method will be auto-detected according
        to extension.

    Returns
    -------
    :class:`bool`
        Definition success.

    References
    ----------
    :cite:`AdobeSystems2013b`, :cite:`Chamberlain2015`,
    :cite:`RisingSunResearch`

    Examples
    --------
    Writing a 3x1D *Iridas* *.cube* *LUT*:

    >>> import numpy as np
    >>> from colour.algebra import spow
    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> LUT = LUT3x1D(
    ...     spow(LUT3x1D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT(LUT, 'My_LUT.cube')  # doctest: +SKIP

    Writing a 1D *Sony* *.spi1d* *LUT*:

    >>> domain = np.array([-0.1, 1.5])
    >>> LUT = LUT1D(
    ...     spow(LUT1D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT(LUT, 'My_LUT.spi1d')  # doctest: +SKIP

    Writing a 3D *Sony* *.spi3d* *LUT*:

    >>> LUT = LUT3D(
    ...     LUT3D.linear_table(16) ** (1 / 2.2),
    ...     'My LUT',
    ...     np.array([[0, 0, 0], [1, 1, 1]]),
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT(LUT, 'My_LUT.cube')  # doctest: +SKIP
    """

    method = cast(
        str,
        optional(
            method, MAPPING_EXTENSION_TO_LUT_FORMAT[os.path.splitext(path)[-1]]
        ),
    )

    method = validate_method(method, LUT_WRITE_METHODS)

    if method == "iridas cube" and isinstance(LUT, LUTSequence):
        method = "resolve cube"

    function = LUT_WRITE_METHODS[method]

    return function(LUT, path, decimals, **filter_kwargs(function, **kwargs))


__all__ += [
    "LUT_READ_METHODS",
    "read_LUT",
    "LUT_WRITE_METHODS",
    "write_LUT",
]



"""
SMPTE ST 2084:2014
==================

Defines the *SMPTE ST 2084:2014* opto-electrical transfer function
(OETF) and electro-optical transfer function (EOTF):

-   :func:`colour.models.eotf_ST2084`
-   :func:`colour.models.eotf_inverse_ST2084`

References
----------
-   :cite:`Miller2014a` : Miller, S. (2014). A Perceptual EOTF for Extended
    Dynamic Range Imagery (pp. 1-17).
    https://www.smpte.org/sites/default/files/\
2014-05-06-EOTF-Miller-1-2-handout.pdf
-   :cite:`SocietyofMotionPictureandTelevisionEngineers2014a` : Society of
    Motion Picture and Television Engineers. (2014). SMPTE ST 2084:2014 -
    Dynamic Range Electro-Optical Transfer Function of Mastering Reference
    Displays (pp. 1-14). doi:10.5594/SMPTE.ST2084.2014
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.hints import Floating, FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import Structure, as_float_array, as_float

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANTS_ST2084",
    "eotf_inverse_ST2084",
    "eotf_ST2084",
]

CONSTANTS_ST2084: Structure = Structure(
    m_1=2610 / 4096 * (1 / 4),
    m_2=2523 / 4096 * 128,
    c_1=3424 / 4096,
    c_2=2413 / 4096 * 32,
    c_3=2392 / 4096 * 32,
)
"""
Constants for *SMPTE ST 2084:2014* inverse electro-optical transfer function
(EOTF) and electro-optical transfer function (EOTF).
"""


def eotf_inverse_ST2084(
    C: FloatingOrArrayLike,
    L_p: Floating = 10000,
    constants: Structure = CONSTANTS_ST2084,
) -> FloatingOrNDArray:
    """
    Define *SMPTE ST 2084:2014* optimised perceptual inverse electro-optical
    transfer function (EOTF).

    Parameters
    ----------
    C
        Target optical output :math:`C` in :math:`cd/m^2` of the ideal
        reference display.
    L_p
        System peak luminance :math:`cd/m^2`, this parameter should stay at its
        default :math:`10000 cd/m^2` value for practical applications. It is
        exposed so that the definition can be used as a fitting function.
    constants
        *SMPTE ST 2084:2014* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Color value abbreviated as :math:`N`, that is directly proportional to
        the encoded signal representation, and which is not directly
        proportional to the optical output of a display device.

    Warnings
    --------
    *SMPTE ST 2084:2014* is an absolute transfer function.

    Notes
    -----
    -   *SMPTE ST 2084:2014* is an absolute transfer function, thus the
        domain and range values for the *Reference* and *1* scales are only
        indicative that the data is not affected by scale transformations.
        The effective domain of *SMPTE ST 2084:2014* inverse electro-optical
        transfer function (EOTF) is [0.0001, 10000].

    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``C``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``N``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Miller2014a`,
    :cite:`SocietyofMotionPictureandTelevisionEngineers2014a`

    Examples
    --------
    >>> eotf_inverse_ST2084(100)  # doctest: +ELLIPSIS
    0.5080784...
    """

    C = as_float_array(C)

    Y_p = spow(C / L_p, constants.m_1)

    N = spow(
        (constants.c_1 + constants.c_2 * Y_p) / (constants.c_3 * Y_p + 1),
        constants.m_2,
    )

    return as_float(N)


def eotf_ST2084(
    N: FloatingOrArrayLike,
    L_p: Floating = 10000,
    constants: Structure = CONSTANTS_ST2084,
) -> FloatingOrNDArray:
    """
    Define *SMPTE ST 2084:2014* optimised perceptual electro-optical transfer
    function (EOTF).

    This perceptual quantizer (PQ) has been modeled by Dolby Laboratories
    using *Barten (1999)* contrast sensitivity function.

    Parameters
    ----------
    N
        Color value abbreviated as :math:`N`, that is directly proportional to
        the encoded signal representation, and which is not directly
        proportional to the optical output of a display device.
    L_p
        System peak luminance :math:`cd/m^2`, this parameter should stay at its
        default :math:`10000 cd/m^2` value for practical applications. It is
        exposed so that the definition can be used as a fitting function.
    constants
        *SMPTE ST 2084:2014* constants.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
          Target optical output :math:`C` in :math:`cd/m^2` of the ideal
          reference display.

    Warnings
    --------
    *SMPTE ST 2084:2014* is an absolute transfer function.

    Notes
    -----
    -   *SMPTE ST 2084:2014* is an absolute transfer function, thus the
        domain and range values for the *Reference* and *1* scales are only
        indicative that the data is not affected by scale transformations.

    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``N``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``C``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Miller2014a`,
    :cite:`SocietyofMotionPictureandTelevisionEngineers2014a`

    Examples
    --------
    >>> eotf_ST2084(0.508078421517399)  # doctest: +ELLIPSIS
    100.0000000...
    """

    N = as_float_array(N)

    m_1_d = 1 / constants.m_1
    m_2_d = 1 / constants.m_2

    V_p = spow(N, m_2_d)

    n = V_p - constants.c_1
    # Limiting negative values.
    n = np.where(n < 0, 0, n)

    L = spow((n / (constants.c_2 - constants.c_3 * V_p)), m_1_d)
    C = L_p * L

    return as_float(C)



"""Showcases *ZCAM* colour appearance model computations."""

import numpy as np

import colour
from colour.appearance.zcam import CAM_ReferenceSpecification_ZCAM
from colour.utilities import message_box

message_box('"ZCAM" Colour Appearance Model Computations')

XYZ = np.array([19.01, 20.00, 21.78])
XYZ_w = np.array([95.05, 100.00, 108.88])
L_A = 318.31
Y_b = 20.0
surround = colour.VIEWING_CONDITIONS_ZCAM["Average"]
message_box(
    f'Converting to the "ZCAM" colour appearance model specification using '
    f"given parameters:\n\n"
    f"\tXYZ: {XYZ}\n"
    f"\tXYZ_w: {XYZ_w}\n"
    f"\tL_A: {L_A}\n"
    f"\tY_b: {Y_b}\n"
    f"\tSurround: {surround}"
)
specification = colour.XYZ_to_ZCAM(XYZ, XYZ_w, L_A, Y_b, surround)
print(specification)

print("\n")

message_box(
    'Broadcasting the current output "ZCAM" colour appearance model '
    "specification to the reference specification."
)

print(CAM_ReferenceSpecification_ZCAM(*specification.values))


print("\n")

J = 48.095883049811491
C = 0.18427174878137914
h = 219.74741565783773
specification = colour.CAM_Specification_ZCAM(J, C, h)
message_box(
    f'Converting to "CIE XYZ" tristimulus values using given parameters:\n\n'
    f"\tJ: {J}\n"
    f"\tC: {C}\n"
    f"\th: {h}\n"
    f"\tXYZ_w: {XYZ_w}\n"
    f"\tL_A: {L_A}\n"
    f"\tY_b: {Y_b}\n"
    f"\tSurround: {surround}"
)
print(colour.ZCAM_to_XYZ(specification, XYZ_w, L_A, Y_b, surround))



import sys

from colour.utilities.deprecation import ModuleAPI, build_API_changes
from colour.utilities.documentation import is_documentation_building

from colour.hints import Any

from .cameras import RGB_CameraSensitivities
from .displays import RGB_DisplayPrimaries
from .datasets import *  # noqa
from . import datasets
from .aces_it import (
    sd_to_aces_relative_exposure_values,
    sd_to_ACES2065_1,
    read_training_data_rawtoaces_v1,
    generate_illuminants_rawtoaces_v1,
    white_balance_multipliers,
    best_illuminant,
    normalise_illuminant,
    training_data_sds_to_RGB,
    training_data_sds_to_XYZ,
    optimisation_factory_rawtoaces_v1,
    optimisation_factory_Jzazbz,
    matrix_idt,
    camera_RGB_to_ACES2065_1,
)
from .correction import (
    matrix_augmented_Cheung2004,
    polynomial_expansion_Finlayson2015,
    polynomial_expansion_Vandermonde,
    POLYNOMIAL_EXPANSION_METHODS,
    polynomial_expansion,
    matrix_colour_correction_Cheung2004,
    matrix_colour_correction_Finlayson2015,
    matrix_colour_correction_Vandermonde,
    MATRIX_COLOUR_CORRECTION_METHODS,
    matrix_colour_correction,
    colour_correction_Cheung2004,
    colour_correction_Finlayson2015,
    colour_correction_Vandermonde,
    COLOUR_CORRECTION_METHODS,
    colour_correction,
)

__all__ = [
    "RGB_CameraSensitivities",
]
__all__ += [
    "RGB_DisplayPrimaries",
]
__all__ += datasets.__all__
__all__ += [
    "sd_to_aces_relative_exposure_values",
    "sd_to_ACES2065_1",
    "read_training_data_rawtoaces_v1",
    "generate_illuminants_rawtoaces_v1",
    "white_balance_multipliers",
    "best_illuminant",
    "normalise_illuminant",
    "training_data_sds_to_RGB",
    "training_data_sds_to_XYZ",
    "optimisation_factory_rawtoaces_v1",
    "optimisation_factory_Jzazbz",
    "matrix_idt",
    "camera_RGB_to_ACES2065_1",
]
__all__ += [
    "matrix_augmented_Cheung2004",
    "polynomial_expansion_Finlayson2015",
    "polynomial_expansion_Vandermonde",
    "POLYNOMIAL_EXPANSION_METHODS",
    "polynomial_expansion",
    "matrix_colour_correction_Cheung2004",
    "matrix_colour_correction_Finlayson2015",
    "matrix_colour_correction_Vandermonde",
    "MATRIX_COLOUR_CORRECTION_METHODS",
    "matrix_colour_correction",
    "colour_correction_Cheung2004",
    "colour_correction_Finlayson2015",
    "colour_correction_Vandermonde",
    "COLOUR_CORRECTION_METHODS",
    "colour_correction",
]


# ----------------------------------------------------------------------------#
# ---                API Changes and Deprecation Management                ---#
# ----------------------------------------------------------------------------#
class characterisation(ModuleAPI):
    """Define a class acting like the *characterisation* module."""

    def __getattr__(self, attribute) -> Any:
        """Return the value from the attribute with given name."""

        return super().__getattr__(attribute)


# v0.4.0
API_CHANGES = {
    "ObjectRenamed": [
        [
            "colour.characterisation.optimisation_factory_JzAzBz",
            "colour.characterisation.optimisation_factory_Jzazbz",
        ],
    ]
}
"""Defines the *colour.characterisation* sub-package API changes."""

if not is_documentation_building():
    sys.modules[
        "colour.characterisation"
    ] = characterisation(  # type: ignore[assignment]
        sys.modules["colour.characterisation"], build_API_changes(API_CHANGES)
    )

    del ModuleAPI, is_documentation_building, build_API_changes, sys



"""
Hexadecimal Notation
====================

Defines the objects for hexadecimal notation:

-   :func:`colour.notation.RGB_to_HEX`
-   :func:`colour.notation.HEX_to_RGB`
"""

from __future__ import annotations

import numpy as np

from colour.algebra import normalise_maximum
from colour.hints import ArrayLike, List, NDArray, StrOrArrayLike, StrOrNDArray
from colour.models import eotf_inverse_sRGB, eotf_sRGB
from colour.utilities import (
    as_float_array,
    as_int_array,
    from_range_1,
    to_domain_1,
    usage_warning,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_to_HEX",
    "HEX_to_RGB",
]


def RGB_to_HEX(RGB: ArrayLike) -> StrOrNDArray:
    """
    Convert from *RGB* colourspace to hexadecimal representation.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.

    Returns
    -------
    :class:`str` or :class:`numpy.array`
        Hexadecimal representation.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> RGB = np.array([0.66666667, 0.86666667, 1.00000000])
    >>> RGB_to_HEX(RGB)
    '#aaddff'
    """

    RGB = to_domain_1(RGB)

    if np.any(RGB < 0):
        usage_warning(
            '"RGB" array contains negative values, those will be clipped, '
            "unpredictable results may occur!"
        )

        RGB = as_float_array(np.clip(RGB, 0, np.inf))

    if np.any(RGB > 1):
        usage_warning(
            '"RGB" array contains values over 1 and will be normalised, '
            "unpredictable results may occur!"
        )

        RGB = eotf_inverse_sRGB(normalise_maximum(eotf_sRGB(RGB)))

    to_HEX = np.vectorize("{:02x}".format)

    HEX = to_HEX(as_int_array(RGB * 255, dtype=np.uint8)).astype(object)
    HEX = np.asarray("#") + HEX[..., 0] + HEX[..., 1] + HEX[..., 2]

    return HEX


def HEX_to_RGB(HEX: StrOrArrayLike) -> NDArray:
    """
    Convert from hexadecimal representation to *RGB* colourspace.

    Parameters
    ----------
    HEX
        Hexadecimal representation.

    Returns
    -------
    :class:`numpy.array`
        *RGB* colourspace array.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``RGB``   | [0, 1]                | [0, 1]        |
    +-----------+-----------------------+---------------+

    Examples
    --------
    >>> HEX = '#aaddff'
    >>> HEX_to_RGB(HEX)  # doctest: +ELLIPSIS
    array([ 0.6666666...,  0.8666666...,  1.        ])
    """

    HEX = np.core.defchararray.lstrip(HEX, "#")  # type: ignore[arg-type]

    def to_RGB(x: List) -> List:
        """Convert given hexadecimal representation to *RGB*."""

        l_x = len(x)

        return [
            int(x[i : i + l_x // 3], 16)  # type: ignore[call-overload]
            for i in range(0, l_x, l_x // 3)
        ]

    to_RGB_v = np.vectorize(to_RGB, otypes=[np.ndarray])

    RGB = as_float_array(to_RGB_v(HEX).tolist()) / 255

    return from_range_1(RGB)



"""Defines the unit tests for the :mod:`colour.models.cam02_ucs` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.appearance import (
    CAM_KWARGS_CIECAM02_sRGB,
    VIEWING_CONDITIONS_CIECAM02,
    XYZ_to_CIECAM02,
)
from colour.models.cam02_ucs import (
    COEFFICIENTS_UCS_LUO2006,
    JMh_CIECAM02_to_UCS_Luo2006,
    UCS_Luo2006_to_JMh_CIECAM02,
    XYZ_to_UCS_Luo2006,
    UCS_Luo2006_to_XYZ,
)
from colour.models import (
    JMh_CIECAM02_to_CAM02LCD,
    CAM02LCD_to_JMh_CIECAM02,
    JMh_CIECAM02_to_CAM02SCD,
    CAM02SCD_to_JMh_CIECAM02,
    JMh_CIECAM02_to_CAM02UCS,
    CAM02UCS_to_JMh_CIECAM02,
    XYZ_to_CAM02LCD,
    CAM02LCD_to_XYZ,
    XYZ_to_CAM02SCD,
    CAM02SCD_to_XYZ,
    XYZ_to_CAM02UCS,
    CAM02UCS_to_XYZ,
)
from colour.utilities import attest, domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestJMh_CIECAM02_to_UCS_Luo2006",
    "TestUCS_Luo2006_to_JMh_CIECAM02",
    "TestXYZ_to_UCS_Luo2006",
    "TestUCS_Luo2006_to_XYZ",
]


class TestJMh_CIECAM02_to_UCS_Luo2006(unittest.TestCase):
    """
    Define :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006`
    definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20.0
        surround = VIEWING_CONDITIONS_CIECAM02["Average"]
        specification = XYZ_to_CIECAM02(XYZ, XYZ_w, L_A, Y_b, surround)

        self._JMh = np.array(
            [specification.J, specification.M, specification.h]
        )

    def test_JMh_CIECAM02_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006`
        definition.
        """

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            np.array([54.90433134, -0.08450395, -0.06854831]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            JMh_CIECAM02_to_CAM02LCD(self._JMh),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-SCD"]
            ),
            np.array([54.90433134, -0.08436178, -0.06843298]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-SCD"]
            ),
            JMh_CIECAM02_to_CAM02SCD(self._JMh),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-UCS"]
            ),
            np.array([54.90433134, -0.08442362, -0.06848314]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                self._JMh, COEFFICIENTS_UCS_LUO2006["CAM02-UCS"]
            ),
            JMh_CIECAM02_to_CAM02UCS(self._JMh),
            decimal=7,
        )

    def test_n_dimensional_JMh_CIECAM02_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006`
        definition n-dimensional support.
        """

        JMh = self._JMh
        Jpapbp = JMh_CIECAM02_to_UCS_Luo2006(
            JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
        )

        JMh = np.tile(JMh, (6, 1))
        Jpapbp = np.tile(Jpapbp, (6, 1))
        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            Jpapbp,
            decimal=7,
        )

        JMh = np.reshape(JMh, (2, 3, 3))
        Jpapbp = np.reshape(Jpapbp, (2, 3, 3))
        np.testing.assert_almost_equal(
            JMh_CIECAM02_to_UCS_Luo2006(
                JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            Jpapbp,
            decimal=7,
        )

    def test_domain_range_scale_JMh_CIECAM02_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006`
        definition domain and range scale support.
        """

        JMh = self._JMh
        Jpapbp = JMh_CIECAM02_to_UCS_Luo2006(
            JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
        )

        d_r = (
            ("reference", 1, 1),
            ("1", np.array([0.01, 0.01, 1 / 360]), 0.01),
            ("100", np.array([1, 1, 1 / 3.6]), 1),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    JMh_CIECAM02_to_UCS_Luo2006(
                        JMh * factor_a, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
                    ),
                    Jpapbp * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_JMh_CIECAM02_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.JMh_CIECAM02_to_UCS_Luo2006`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            JMh = np.array(case)
            JMh_CIECAM02_to_UCS_Luo2006(
                JMh, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            )


class TestUCS_Luo2006_to_JMh_CIECAM02(unittest.TestCase):
    """
    Define :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02`
    definition unit tests methods.
    """

    def test_UCS_Luo2006_to_JMh_CIECAM02(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02`
        definition.
        """

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            np.array([41.73109113, 0.10873867, 219.04843202]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            CAM02LCD_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314])
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            np.array([41.73109113, 0.10892212, 219.04843202]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            CAM02SCD_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314])
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            np.array([41.73109113, 0.10884218, 219.04843202]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            CAM02UCS_to_JMh_CIECAM02(
                np.array([54.90433134, -0.08442362, -0.06848314])
            ),
            decimal=7,
        )

    def test_n_dimensional_UCS_Luo2006_to_JMh_CIECAM02(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02`
        definition n-dimensional support.
        """

        Jpapbp = np.array([54.90433134, -0.08442362, -0.06848314])
        JMh = UCS_Luo2006_to_JMh_CIECAM02(
            Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
        )

        Jpapbp = np.tile(Jpapbp, (6, 1))
        JMh = np.tile(JMh, (6, 1))
        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            JMh,
            decimal=7,
        )

        Jpapbp = np.reshape(Jpapbp, (2, 3, 3))
        JMh = np.reshape(JMh, (2, 3, 3))
        np.testing.assert_almost_equal(
            UCS_Luo2006_to_JMh_CIECAM02(
                Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            ),
            JMh,
            decimal=7,
        )

    def test_domain_range_scale_UCS_Luo2006_to_JMh_CIECAM02(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02`
        definition domain and range scale support.
        """

        Jpapbp = np.array([54.90433134, -0.08442362, -0.06848314])
        JMh = UCS_Luo2006_to_JMh_CIECAM02(
            Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
        )

        d_r = (
            ("reference", 1, 1),
            ("1", 0.01, np.array([0.01, 0.01, 1 / 360])),
            ("100", 1, np.array([1, 1, 1 / 3.6])),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    UCS_Luo2006_to_JMh_CIECAM02(
                        Jpapbp * factor_a,
                        COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
                    ),
                    JMh * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_UCS_Luo2006_to_JMh_CIECAM02(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_JMh_CIECAM02`
        definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Jpapbp = np.array(case)
            UCS_Luo2006_to_JMh_CIECAM02(
                Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
            )


class TestXYZ_to_UCS_Luo2006(unittest.TestCase):
    """
    Define :func:`colour.models.cam02_ucs.XYZ_to_UCS_Luo2006` definition
    unit tests methods.
    """

    def test_XYZ_to_UCS_Luo2006(self):
        """Test :func:`colour.models.cam02_ucs.XYZ_to_UCS_Luo2006` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            np.array([46.61386154, 39.35760236, 15.96730435]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            XYZ_to_CAM02LCD(np.array([0.20654008, 0.12197225, 0.05136952])),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            np.array([46.61386154, 25.62879882, 10.39755489]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            XYZ_to_CAM02SCD(np.array([0.20654008, 0.12197225, 0.05136952])),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            np.array([46.61386154, 29.88310013, 12.12351683]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            XYZ_to_CAM02UCS(np.array([0.20654008, 0.12197225, 0.05136952])),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.XYZ_to_UCS_Luo2006` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Jpapbp = XYZ_to_UCS_Luo2006(XYZ, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"])

        XYZ = np.tile(XYZ, (6, 1))
        Jpapbp = np.tile(Jpapbp, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(XYZ, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]),
            Jpapbp,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        Jpapbp = np.reshape(Jpapbp, (2, 3, 3))
        np.testing.assert_almost_equal(
            XYZ_to_UCS_Luo2006(XYZ, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]),
            Jpapbp,
            decimal=7,
        )

    def test_domain_range_scale_XYZ_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.XYZ_to_UCS_Luo2006` definition
        domain and range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        XYZ_w = CAM_KWARGS_CIECAM02_sRGB["XYZ_w"] / 100
        Jpapbp = XYZ_to_UCS_Luo2006(XYZ, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"])

        d_r = (("reference", 1, 1), ("1", 1, 0.01), ("100", 100, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_UCS_Luo2006(
                        XYZ * factor_a,
                        COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
                        XYZ_w=XYZ_w * factor_a,
                    ),
                    Jpapbp * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_UCS_Luo2006(self):
        """
        Test :func:`colour.models.cam02_ucs.XYZ_to_UCS_Luo2006` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_to_UCS_Luo2006(XYZ, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"])


class TestUCS_Luo2006_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.cam02_ucs.UCS_Luo2006_to_XYZ` definition
    unit tests methods.
    """

    def test_UCS_Luo2006_to_XYZ(self):
        """Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_XYZ` definition."""

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
            ),
            CAM02LCD_to_XYZ(np.array([46.61386154, 39.35760236, 15.96730435])),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            np.array([0.28264475, 0.11036927, 0.00824593]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-SCD"],
            ),
            CAM02SCD_to_XYZ(np.array([46.61386154, 39.35760236, 15.96730435])),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            np.array([0.24229809, 0.11573005, 0.02517649]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(
                np.array([46.61386154, 39.35760236, 15.96730435]),
                COEFFICIENTS_UCS_LUO2006["CAM02-UCS"],
            ),
            CAM02UCS_to_XYZ(np.array([46.61386154, 39.35760236, 15.96730435])),
            decimal=7,
        )

    def test_n_dimensional_UCS_Luo2006_to_XYZ(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_XYZ` definition
        n-dimensional support.
        """

        Jpapbp = np.array([46.61386154, 39.35760236, 15.96730435])
        XYZ = UCS_Luo2006_to_XYZ(Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"])

        Jpapbp = np.tile(Jpapbp, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]),
            XYZ,
            decimal=7,
        )

        Jpapbp = np.reshape(Jpapbp, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            UCS_Luo2006_to_XYZ(Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]),
            XYZ,
            decimal=7,
        )

    def test_domain_range_scale_UCS_Luo2006_to_XYZ(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_XYZ` definition
        domain and range scale support.
        """

        Jpapbp = np.array([46.61386154, 39.35760236, 15.96730435])
        XYZ = UCS_Luo2006_to_XYZ(Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"])
        XYZ_w = CAM_KWARGS_CIECAM02_sRGB["XYZ_w"] / 100

        d_r = (("reference", 1, 1, 1), ("1", 0.01, 1, 1), ("100", 1, 100, 100))
        for scale, factor_a, factor_b, factor_c in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    UCS_Luo2006_to_XYZ(
                        Jpapbp * factor_a,
                        COEFFICIENTS_UCS_LUO2006["CAM02-LCD"],
                        XYZ_w=XYZ_w * factor_c,
                    ),
                    XYZ * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_UCS_Luo2006_to_XYZ(self):
        """
        Test :func:`colour.models.cam02_ucs.UCS_Luo2006_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Jpapbp = np.array(case)
            try:
                UCS_Luo2006_to_XYZ(
                    Jpapbp, COEFFICIENTS_UCS_LUO2006["CAM02-LCD"]
                )
            except ValueError as error:
                attest("CAM_Specification_CIECAM02" in str(error))


if __name__ == "__main__":
    unittest.main()
